# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest


from unittest import mock
from importlib import reload
from unittest.mock import patch
from unittest.mock import ANY

from google.api_core import exceptions
from google.api_core import operation

from google.cloud import aiplatform
from google.cloud.aiplatform.vizier import Study
from google.cloud.aiplatform.vizier import Trial
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform.vizier import pyvizier

from google.cloud.aiplatform.compat.services import vizier_service_client
from google.cloud.aiplatform.compat.types import (
    study as gca_study,
    vizier_service as gca_vizier_service,
)
from google.protobuf import duration_pb2


# project
_TEST_PROJECT = "test-project"
_TEST_LOCATION = "us-central1"
_TEST_PARENT = f"projects/{_TEST_PROJECT}/locations/{_TEST_LOCATION}"

# study
_TEST_STUDY_ID = "12345"
_TEST_STUDY_NAME = f"{_TEST_PARENT}/studies/{_TEST_STUDY_ID}"

# trial
_TEST_TRIAL_ID = "1"
_TEST_TRIAL_NAME = f"{_TEST_STUDY_NAME}/trials/{_TEST_TRIAL_ID}"

_TEST_METRIC_ID = "pr-auc"
_TEST_DISPLAY_NAME = "test_study_python_aiplatform"

_TEST_PARAMETER_ID_1 = "learning_rate"
_TEST_PARAMETER_ID_MIN_VALUE_1 = 1e-05
_TEST_PARAMETER_ID_MAX_VALUE_1 = 1.0

_TEST_PARAMETER_ID_2 = "optimizer"
_TEST_PARAMETER_VALUE_2 = ["adagrad", "adam", "experimental"]

_TEST_STUDY = gca_study.Study(
    display_name=_TEST_DISPLAY_NAME,
    study_spec=gca_study.StudySpec(
        algorithm=gca_study.StudySpec.Algorithm.RANDOM_SEARCH,
        metrics=[
            gca_study.StudySpec.MetricSpec(
                metric_id=_TEST_METRIC_ID,
                goal=gca_study.StudySpec.MetricSpec.GoalType.MAXIMIZE,
            )
        ],
        parameters=[
            gca_study.StudySpec.ParameterSpec(
                parameter_id=_TEST_PARAMETER_ID_1,
                scale_type=gca_study.StudySpec.ParameterSpec.ScaleType.UNIT_LINEAR_SCALE,
                double_value_spec=gca_study.StudySpec.ParameterSpec.DoubleValueSpec(
                    min_value=_TEST_PARAMETER_ID_MIN_VALUE_1,
                    max_value=_TEST_PARAMETER_ID_MAX_VALUE_1,
                ),
            ),
            gca_study.StudySpec.ParameterSpec(
                parameter_id=_TEST_PARAMETER_ID_2,
                categorical_value_spec=gca_study.StudySpec.ParameterSpec.CategoricalValueSpec(
                    values=_TEST_PARAMETER_VALUE_2
                ),
            ),
        ],
    ),
)


@pytest.fixture
def get_study_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "get_study"
    ) as get_study_mock:
        get_study_mock.return_value = gca_study.Study(name=_TEST_STUDY_NAME)
        yield get_study_mock


@pytest.fixture
def get_trial_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "get_trial"
    ) as get_trial_mock:
        get_trial_mock.return_value = gca_study.Trial(
            name=_TEST_TRIAL_NAME,
            state=gca_study.Trial.State.ACTIVE,
            parameters=[
                gca_study.Trial.Parameter(
                    parameter_id=_TEST_PARAMETER_ID_1,
                    value=_TEST_PARAMETER_ID_MIN_VALUE_1,
                )
            ],
        )
        yield get_trial_mock


@pytest.fixture
def create_study_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "create_study"
    ) as create_study_mock:
        create_study_mock.return_value = gca_study.Study(
            name=_TEST_STUDY_NAME,
        )
        yield create_study_mock


@pytest.fixture
def lookup_study_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "lookup_study"
    ) as lookup_study_mock:
        lookup_study_mock.return_value = gca_study.Study(
            name=_TEST_STUDY_NAME,
        )
        yield lookup_study_mock


@pytest.fixture
def suggest_trials_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "suggest_trials"
    ) as suggest_trials_mock:
        suggest_trials_lro_mock = mock.Mock(operation.Operation)
        suggest_trials_lro_mock.result.return_value = (
            gca_vizier_service.SuggestTrialsResponse(
                trials=[gca_study.Trial(name=_TEST_TRIAL_NAME)]
            )
        )
        suggest_trials_mock.return_value = suggest_trials_lro_mock
        yield suggest_trials_mock


@pytest.fixture
def list_optimal_trials_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "list_optimal_trials"
    ) as list_optimal_trials_mock:
        list_optimal_trials_mock.return_value = (
            gca_vizier_service.ListOptimalTrialsResponse(
                optimal_trials=[gca_study.Trial(name=_TEST_TRIAL_NAME)]
            )
        )
        yield list_optimal_trials_mock


@pytest.fixture
def list_trials_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "list_trials"
    ) as list_trials_mock:
        list_trials_mock.return_value = gca_vizier_service.ListTrialsResponse(
            trials=[gca_study.Trial(name=_TEST_TRIAL_NAME)]
        )
        yield list_trials_mock


@pytest.fixture
def delete_study_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "delete_study"
    ) as delete_study_mock:
        yield delete_study_mock


@pytest.fixture
def delete_trial_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "delete_trial"
    ) as delete_trial_mock:
        yield delete_trial_mock


@pytest.fixture
def complete_trial_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "complete_trial"
    ) as complete_trial_mock:
        complete_trial_mock.return_value = gca_study.Trial(
            name=_TEST_TRIAL_NAME,
            final_measurement=gca_study.Measurement(
                step_count=3,
                metrics=[gca_study.Measurement.Metric(metric_id="y", value=5)],
            ),
        )
        yield complete_trial_mock


@pytest.fixture
def complete_trial_empty_measurement_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "complete_trial"
    ) as complete_trial_empty_measurement_mock:
        complete_trial_empty_measurement_mock.return_value = gca_study.Trial(
            name=_TEST_TRIAL_NAME
        )
        yield complete_trial_empty_measurement_mock


@pytest.fixture
def should_stop_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "check_trial_early_stopping_state"
    ) as should_stop_mock:
        should_stop_lro_mock = mock.Mock(operation.Operation)
        should_stop_lro_mock.result.return_value = (
            gca_vizier_service.CheckTrialEarlyStoppingStateResponse(should_stop=True)
        )
        should_stop_mock.return_value = should_stop_lro_mock
        yield should_stop_mock


@pytest.fixture
def create_study_mock_already_exists():
    with patch.object(
        vizier_service_client.VizierServiceClient, "create_study"
    ) as create_study_mock:
        create_study_mock.side_effect = [
            exceptions.AlreadyExists("Study already exists."),
            gca_study.Study(
                name=_TEST_STUDY_NAME,
            ),
        ]
        yield create_study_mock


@pytest.fixture
def add_measurement_mock():
    with patch.object(
        vizier_service_client.VizierServiceClient, "add_trial_measurement"
    ) as add_measurement_mock:
        yield add_measurement_mock


@pytest.mark.usefixtures("google_auth_mock")
class TestStudy:
    def setup_method(self):
        reload(initializer)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    @pytest.mark.usefixtures("get_study_mock")
    def test_create_study(self, create_study_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)

        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        create_study_mock.assert_called_once_with(
            parent=_TEST_PARENT, study=_TEST_STUDY
        )
        assert type(study) == Study

    @pytest.mark.usefixtures("get_study_mock")
    def test_create_study_already_exists(
        self, create_study_mock_already_exists, lookup_study_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)

        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        lookup_study_mock.assert_called_once_with(
            request={"parent": _TEST_PARENT, "display_name": _TEST_DISPLAY_NAME}
        )
        assert type(study) == Study

    @pytest.mark.usefixtures("get_study_mock")
    def test_materialize_study_config(self, create_study_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)
        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        study_config = study.materialize_study_config()

        create_study_mock.assert_called_once_with(
            parent=_TEST_PARENT, study=_TEST_STUDY
        )
        assert type(study_config) == pyvizier.StudyConfig

    @pytest.mark.usefixtures("get_study_mock", "get_trial_mock")
    def test_suggest(self, create_study_mock, suggest_trials_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)
        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        trials = study.suggest(count=5, worker="test_worker")

        suggest_trials_mock.assert_called_once_with(
            request={
                "parent": _TEST_STUDY_NAME,
                "suggestion_count": 5,
                "client_id": "test_worker",
            }
        )
        assert type(trials[0]) == Trial

    @pytest.mark.usefixtures("get_study_mock")
    def test_from_uid(self):
        aiplatform.init(project=_TEST_PROJECT)

        study = Study.from_uid(uid=_TEST_STUDY_ID)

        assert type(study) == Study
        assert study.name == _TEST_STUDY_ID

    @pytest.mark.usefixtures("get_study_mock")
    def test_delete(self, create_study_mock, delete_study_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)
        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        study.delete()

        delete_study_mock.assert_called_once_with(name=_TEST_STUDY_NAME)

    @pytest.mark.usefixtures("get_study_mock", "create_study_mock", "get_trial_mock")
    def test_optimal_trials(self, list_optimal_trials_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)
        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        trials = study.optimal_trials()

        list_optimal_trials_mock.assert_called_once_with(
            request={"parent": _TEST_STUDY_NAME}
        )
        assert type(trials[0]) == Trial

    @pytest.mark.usefixtures("get_study_mock", "create_study_mock", "get_trial_mock")
    def test_list_trials(self, list_trials_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)
        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        trials = study.trials()

        list_trials_mock.assert_called_once_with(request={"parent": _TEST_STUDY_NAME})
        assert type(trials[0]) == Trial

    @pytest.mark.usefixtures("get_study_mock", "create_study_mock")
    def test_get_trial(self, get_trial_mock):
        aiplatform.init(project=_TEST_PROJECT)
        sc = pyvizier.StudyConfig()
        sc.algorithm = pyvizier.Algorithm.RANDOM_SEARCH
        sc.metric_information.append(
            pyvizier.MetricInformation(
                name=_TEST_METRIC_ID, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
            )
        )
        root = sc.search_space.select_root()
        root.add_float_param(
            _TEST_PARAMETER_ID_1,
            _TEST_PARAMETER_ID_MIN_VALUE_1,
            _TEST_PARAMETER_ID_MAX_VALUE_1,
            scale_type=pyvizier.ScaleType.LINEAR,
        )
        root.add_categorical_param(_TEST_PARAMETER_ID_2, _TEST_PARAMETER_VALUE_2)
        study = Study.create_or_load(display_name=_TEST_DISPLAY_NAME, problem=sc)

        trial = study.get_trial(1)

        get_trial_mock.assert_called_once_with(name=_TEST_TRIAL_NAME, retry=ANY)
        assert type(trial) == Trial


@pytest.mark.usefixtures("google_auth_mock")
class TestTrial:
    def setup_method(self):
        reload(initializer)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    @pytest.mark.usefixtures("get_trial_mock")
    def test_delete(self, delete_trial_mock):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)

        trial.delete()

        delete_trial_mock.assert_called_once_with(name=_TEST_TRIAL_NAME)
        assert type(trial) == Trial

    @pytest.mark.usefixtures("get_trial_mock")
    def test_complete(self, complete_trial_mock):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)
        measurement = pyvizier.Measurement()
        measurement.metrics["y"] = 4

        measurement = trial.complete(
            measurement=measurement, infeasible_reason="infeasible"
        )

        complete_trial_mock.assert_called_once_with(
            request={
                "name": _TEST_TRIAL_NAME,
                "infeasible_reason": "infeasible",
                "trial_infeasible": True,
                "final_measurement": gca_study.Measurement(
                    elapsed_duration=duration_pb2.Duration(),
                    metrics=[gca_study.Measurement.Metric(metric_id="y", value=4)],
                ),
            }
        )
        assert type(measurement) == pyvizier.Measurement

    @pytest.mark.usefixtures("get_trial_mock")
    def test_complete_empty_measurement(self, complete_trial_empty_measurement_mock):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)
        measurement = pyvizier.Measurement()
        measurement.metrics["y"] = 4

        measurement = trial.complete(
            measurement=measurement, infeasible_reason="infeasible"
        )

        complete_trial_empty_measurement_mock.assert_called_once_with(
            request={
                "name": _TEST_TRIAL_NAME,
                "infeasible_reason": "infeasible",
                "trial_infeasible": True,
                "final_measurement": gca_study.Measurement(
                    elapsed_duration=duration_pb2.Duration(),
                    metrics=[gca_study.Measurement.Metric(metric_id="y", value=4)],
                ),
            }
        )
        assert measurement is None

    @pytest.mark.usefixtures("get_trial_mock")
    def test_should_stop(self, should_stop_mock):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)

        should_stop = trial.should_stop()

        should_stop_mock.assert_called_once_with(
            request={"trial_name": _TEST_TRIAL_NAME}
        )
        assert should_stop is True

    @pytest.mark.usefixtures("get_trial_mock")
    def test_add_measurement(self, add_measurement_mock):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)
        measurement = pyvizier.Measurement()
        measurement.metrics["y"] = 4

        add_measurement = trial.add_measurement(measurement=measurement)

        add_measurement_mock.assert_called_once_with(
            request={
                "trial_name": _TEST_TRIAL_NAME,
                "measurement": gca_study.Measurement(
                    elapsed_duration=duration_pb2.Duration(),
                    metrics=[gca_study.Measurement.Metric(metric_id="y", value=4)],
                ),
            }
        )
        assert add_measurement is None

    @pytest.mark.usefixtures("get_trial_mock")
    def test_properties(self):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)
        measurement = pyvizier.Measurement()
        measurement.metrics["y"] = 4

        uid = trial.uid
        status = trial.status
        parameters = trial.parameters

        assert uid == 1
        assert status == pyvizier.TrialStatus.ACTIVE
        assert (
            parameters.get_value(_TEST_PARAMETER_ID_1) == _TEST_PARAMETER_ID_MIN_VALUE_1
        )

    @pytest.mark.usefixtures("get_trial_mock")
    def test_materialize(self):
        aiplatform.init(project=_TEST_PROJECT)
        trial = Trial(trial_name=_TEST_TRIAL_NAME)
        measurement = pyvizier.Measurement()
        measurement.metrics["y"] = 4

        materialize_trial = trial.materialize()

        assert materialize_trial.id == 1
        assert (
            materialize_trial.parameters.get_value(_TEST_PARAMETER_ID_1)
            == _TEST_PARAMETER_ID_MIN_VALUE_1
        )



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import index as gca_index
from google.cloud.aiplatform_v1.types import operation
from google.protobuf import field_mask_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "CreateIndexRequest",
        "CreateIndexOperationMetadata",
        "GetIndexRequest",
        "ListIndexesRequest",
        "ListIndexesResponse",
        "UpdateIndexRequest",
        "UpdateIndexOperationMetadata",
        "DeleteIndexRequest",
        "UpsertDatapointsRequest",
        "UpsertDatapointsResponse",
        "RemoveDatapointsRequest",
        "RemoveDatapointsResponse",
        "NearestNeighborSearchOperationMetadata",
    },
)


class CreateIndexRequest(proto.Message):
    r"""Request message for
    [IndexService.CreateIndex][google.cloud.aiplatform.v1.IndexService.CreateIndex].

    Attributes:
        parent (str):
            Required. The resource name of the Location to create the
            Index in. Format:
            ``projects/{project}/locations/{location}``
        index (google.cloud.aiplatform_v1.types.Index):
            Required. The Index to create.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    index: gca_index.Index = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gca_index.Index,
    )


class CreateIndexOperationMetadata(proto.Message):
    r"""Runtime operation information for
    [IndexService.CreateIndex][google.cloud.aiplatform.v1.IndexService.CreateIndex].

    Attributes:
        generic_metadata (google.cloud.aiplatform_v1.types.GenericOperationMetadata):
            The operation generic information.
        nearest_neighbor_search_operation_metadata (google.cloud.aiplatform_v1.types.NearestNeighborSearchOperationMetadata):
            The operation metadata with regard to
            Matching Engine Index operation.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )
    nearest_neighbor_search_operation_metadata: "NearestNeighborSearchOperationMetadata" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="NearestNeighborSearchOperationMetadata",
    )


class GetIndexRequest(proto.Message):
    r"""Request message for
    [IndexService.GetIndex][google.cloud.aiplatform.v1.IndexService.GetIndex]

    Attributes:
        name (str):
            Required. The name of the Index resource. Format:
            ``projects/{project}/locations/{location}/indexes/{index}``
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class ListIndexesRequest(proto.Message):
    r"""Request message for
    [IndexService.ListIndexes][google.cloud.aiplatform.v1.IndexService.ListIndexes].

    Attributes:
        parent (str):
            Required. The resource name of the Location from which to
            list the Indexes. Format:
            ``projects/{project}/locations/{location}``
        filter (str):
            The standard list filter.
        page_size (int):
            The standard list page size.
        page_token (str):
            The standard list page token. Typically obtained via
            [ListIndexesResponse.next_page_token][google.cloud.aiplatform.v1.ListIndexesResponse.next_page_token]
            of the previous
            [IndexService.ListIndexes][google.cloud.aiplatform.v1.IndexService.ListIndexes]
            call.
        read_mask (google.protobuf.field_mask_pb2.FieldMask):
            Mask specifying which fields to read.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=2,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=3,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=4,
    )
    read_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=5,
        message=field_mask_pb2.FieldMask,
    )


class ListIndexesResponse(proto.Message):
    r"""Response message for
    [IndexService.ListIndexes][google.cloud.aiplatform.v1.IndexService.ListIndexes].

    Attributes:
        indexes (MutableSequence[google.cloud.aiplatform_v1.types.Index]):
            List of indexes in the requested page.
        next_page_token (str):
            A token to retrieve next page of results. Pass to
            [ListIndexesRequest.page_token][google.cloud.aiplatform.v1.ListIndexesRequest.page_token]
            to obtain that page.
    """

    @property
    def raw_page(self):
        return self

    indexes: MutableSequence[gca_index.Index] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=gca_index.Index,
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


class UpdateIndexRequest(proto.Message):
    r"""Request message for
    [IndexService.UpdateIndex][google.cloud.aiplatform.v1.IndexService.UpdateIndex].

    Attributes:
        index (google.cloud.aiplatform_v1.types.Index):
            Required. The Index which updates the
            resource on the server.
        update_mask (google.protobuf.field_mask_pb2.FieldMask):
            The update mask applies to the resource. For the
            ``FieldMask`` definition, see
            [google.protobuf.FieldMask][google.protobuf.FieldMask].
    """

    index: gca_index.Index = proto.Field(
        proto.MESSAGE,
        number=1,
        message=gca_index.Index,
    )
    update_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=2,
        message=field_mask_pb2.FieldMask,
    )


class UpdateIndexOperationMetadata(proto.Message):
    r"""Runtime operation information for
    [IndexService.UpdateIndex][google.cloud.aiplatform.v1.IndexService.UpdateIndex].

    Attributes:
        generic_metadata (google.cloud.aiplatform_v1.types.GenericOperationMetadata):
            The operation generic information.
        nearest_neighbor_search_operation_metadata (google.cloud.aiplatform_v1.types.NearestNeighborSearchOperationMetadata):
            The operation metadata with regard to
            Matching Engine Index operation.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )
    nearest_neighbor_search_operation_metadata: "NearestNeighborSearchOperationMetadata" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="NearestNeighborSearchOperationMetadata",
    )


class DeleteIndexRequest(proto.Message):
    r"""Request message for
    [IndexService.DeleteIndex][google.cloud.aiplatform.v1.IndexService.DeleteIndex].

    Attributes:
        name (str):
            Required. The name of the Index resource to be deleted.
            Format:
            ``projects/{project}/locations/{location}/indexes/{index}``
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class UpsertDatapointsRequest(proto.Message):
    r"""Request message for
    [IndexService.UpsertDatapoints][google.cloud.aiplatform.v1.IndexService.UpsertDatapoints]

    Attributes:
        index (str):
            Required. The name of the Index resource to be updated.
            Format:
            ``projects/{project}/locations/{location}/indexes/{index}``
        datapoints (MutableSequence[google.cloud.aiplatform_v1.types.IndexDatapoint]):
            A list of datapoints to be created/updated.
    """

    index: str = proto.Field(
        proto.STRING,
        number=1,
    )
    datapoints: MutableSequence[gca_index.IndexDatapoint] = proto.RepeatedField(
        proto.MESSAGE,
        number=2,
        message=gca_index.IndexDatapoint,
    )


class UpsertDatapointsResponse(proto.Message):
    r"""Response message for
    [IndexService.UpsertDatapoints][google.cloud.aiplatform.v1.IndexService.UpsertDatapoints]

    """


class RemoveDatapointsRequest(proto.Message):
    r"""Request message for
    [IndexService.RemoveDatapoints][google.cloud.aiplatform.v1.IndexService.RemoveDatapoints]

    Attributes:
        index (str):
            Required. The name of the Index resource to be updated.
            Format:
            ``projects/{project}/locations/{location}/indexes/{index}``
        datapoint_ids (MutableSequence[str]):
            A list of datapoint ids to be deleted.
    """

    index: str = proto.Field(
        proto.STRING,
        number=1,
    )
    datapoint_ids: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=2,
    )


class RemoveDatapointsResponse(proto.Message):
    r"""Response message for
    [IndexService.RemoveDatapoints][google.cloud.aiplatform.v1.IndexService.RemoveDatapoints]

    """


class NearestNeighborSearchOperationMetadata(proto.Message):
    r"""Runtime operation metadata with regard to Matching Engine
    Index.

    Attributes:
        content_validation_stats (MutableSequence[google.cloud.aiplatform_v1.types.NearestNeighborSearchOperationMetadata.ContentValidationStats]):
            The validation stats of the content (per file) to be
            inserted or updated on the Matching Engine Index resource.
            Populated if contentsDeltaUri is provided as part of
            [Index.metadata][google.cloud.aiplatform.v1.Index.metadata].
            Please note that, currently for those files that are broken
            or has unsupported file format, we will not have the stats
            for those files.
        data_bytes_count (int):
            The ingested data size in bytes.
    """

    class RecordError(proto.Message):
        r"""

        Attributes:
            error_type (google.cloud.aiplatform_v1.types.NearestNeighborSearchOperationMetadata.RecordError.RecordErrorType):
                The error type of this record.
            error_message (str):
                A human-readable message that is shown to the user to help
                them fix the error. Note that this message may change from
                time to time, your code should check against error_type as
                the source of truth.
            source_gcs_uri (str):
                Cloud Storage URI pointing to the original
                file in user's bucket.
            embedding_id (str):
                Empty if the embedding id is failed to parse.
            raw_record (str):
                The original content of this record.
        """

        class RecordErrorType(proto.Enum):
            r"""

            Values:
                ERROR_TYPE_UNSPECIFIED (0):
                    Default, shall not be used.
                EMPTY_LINE (1):
                    The record is empty.
                INVALID_JSON_SYNTAX (2):
                    Invalid json format.
                INVALID_CSV_SYNTAX (3):
                    Invalid csv format.
                INVALID_AVRO_SYNTAX (4):
                    Invalid avro format.
                INVALID_EMBEDDING_ID (5):
                    The embedding id is not valid.
                EMBEDDING_SIZE_MISMATCH (6):
                    The size of the embedding vectors does not
                    match with the specified dimension.
                NAMESPACE_MISSING (7):
                    The ``namespace`` field is missing.
            """
            ERROR_TYPE_UNSPECIFIED = 0
            EMPTY_LINE = 1
            INVALID_JSON_SYNTAX = 2
            INVALID_CSV_SYNTAX = 3
            INVALID_AVRO_SYNTAX = 4
            INVALID_EMBEDDING_ID = 5
            EMBEDDING_SIZE_MISMATCH = 6
            NAMESPACE_MISSING = 7

        error_type: "NearestNeighborSearchOperationMetadata.RecordError.RecordErrorType" = proto.Field(
            proto.ENUM,
            number=1,
            enum="NearestNeighborSearchOperationMetadata.RecordError.RecordErrorType",
        )
        error_message: str = proto.Field(
            proto.STRING,
            number=2,
        )
        source_gcs_uri: str = proto.Field(
            proto.STRING,
            number=3,
        )
        embedding_id: str = proto.Field(
            proto.STRING,
            number=4,
        )
        raw_record: str = proto.Field(
            proto.STRING,
            number=5,
        )

    class ContentValidationStats(proto.Message):
        r"""

        Attributes:
            source_gcs_uri (str):
                Cloud Storage URI pointing to the original
                file in user's bucket.
            valid_record_count (int):
                Number of records in this file that were
                successfully processed.
            invalid_record_count (int):
                Number of records in this file we skipped due
                to validate errors.
            partial_errors (MutableSequence[google.cloud.aiplatform_v1.types.NearestNeighborSearchOperationMetadata.RecordError]):
                The detail information of the partial
                failures encountered for those invalid records
                that couldn't be parsed. Up to 50 partial errors
                will be reported.
        """

        source_gcs_uri: str = proto.Field(
            proto.STRING,
            number=1,
        )
        valid_record_count: int = proto.Field(
            proto.INT64,
            number=2,
        )
        invalid_record_count: int = proto.Field(
            proto.INT64,
            number=3,
        )
        partial_errors: MutableSequence[
            "NearestNeighborSearchOperationMetadata.RecordError"
        ] = proto.RepeatedField(
            proto.MESSAGE,
            number=4,
            message="NearestNeighborSearchOperationMetadata.RecordError",
        )

    content_validation_stats: MutableSequence[
        ContentValidationStats
    ] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=ContentValidationStats,
    )
    data_bytes_count: int = proto.Field(
        proto.INT64,
        number=2,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from unittest import mock
from unittest.mock import patch
from importlib import reload

from google.api_core import operation

from google.cloud import aiplatform
from google.cloud.aiplatform import base
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform import tensorboard
from google.cloud.aiplatform import utils

from google.cloud.aiplatform.compat.services import (
    tensorboard_service_client,
)

from google.cloud.aiplatform.compat.types import (
    encryption_spec as gca_encryption_spec,
    tensorboard as gca_tensorboard,
    tensorboard_data as gca_tensorboard_data,
    tensorboard_experiment as gca_tensorboard_experiment,
    tensorboard_run as gca_tensorboard_run,
    tensorboard_service as gca_tensorboard_service,
    tensorboard_time_series as gca_tensorboard_time_series,
)

from google.protobuf import field_mask_pb2

# project
_TEST_PROJECT = "test-project"
_TEST_LOCATION = "us-central1"
_TEST_PARENT = f"projects/{_TEST_PROJECT}/locations/{_TEST_LOCATION}"
_TEST_ALT_PROJECT = "test-project_alt"

_TEST_ALT_LOCATION = "europe-west4"
_TEST_INVALID_LOCATION = "us-central2"

# tensorboard
_TEST_ID = "1028944691210842416"
_TEST_DISPLAY_NAME = "my_tensorboard_1234"
_TEST_DISPLAY_NAME_UPDATE = "my_tensorboard_1234_update"

_TEST_NAME = (
    f"projects/{_TEST_PROJECT}/locations/{_TEST_LOCATION}/tensorboards/{_TEST_ID}"
)
_TEST_ALT_NAME = (
    f"projects/{_TEST_PROJECT}/locations/{_TEST_ALT_LOCATION}/tensorboards/{_TEST_ID}"
)
_TEST_INVALID_NAME = f"prj/{_TEST_PROJECT}/locations/{_TEST_LOCATION}/{_TEST_ID}"

_TEST_TENSORBOARD_EXPERIMENT_ID = "test-experiment"
_TEST_TENSORBOARD_EXPERIMENT_NAME = (
    f"{_TEST_NAME}/experiments/{_TEST_TENSORBOARD_EXPERIMENT_ID}"
)

_TEST_TENSORBOARD_RUN_ID = "test-run"
_TEST_TENSORBOARD_RUN_NAME = (
    f"{_TEST_TENSORBOARD_EXPERIMENT_NAME}/runs/{_TEST_TENSORBOARD_RUN_ID}"
)

_TEST_TENSORBOARD_TIME_SERIES_ID = "test-time-series"
_TEST_TENSORBOARD_TIME_SERIES_NAME = (
    f"{_TEST_TENSORBOARD_RUN_NAME}/timeSeries/{_TEST_TENSORBOARD_TIME_SERIES_ID}"
)

# request_metadata
_TEST_REQUEST_METADATA = ()

# CMEK encryption
_TEST_ENCRYPTION_KEY_NAME = "key_1234"
_TEST_ENCRYPTION_SPEC = gca_encryption_spec.EncryptionSpec(
    kms_key_name=_TEST_ENCRYPTION_KEY_NAME
)

_TEST_TIME_SERIES_DISPLAY_NAME = "accuracy"


@pytest.fixture
def get_tensorboard_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient, "get_tensorboard"
    ) as get_tensorboard_mock:
        get_tensorboard_mock.return_value = gca_tensorboard.Tensorboard(
            name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )
        yield get_tensorboard_mock


@pytest.fixture
def create_tensorboard_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient, "create_tensorboard"
    ) as create_tensorboard_mock:
        create_tensorboard_lro_mock = mock.Mock(operation.Operation)
        create_tensorboard_lro_mock.result.return_value = gca_tensorboard.Tensorboard(
            name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )
        create_tensorboard_mock.return_value = create_tensorboard_lro_mock
        yield create_tensorboard_mock


@pytest.fixture
def update_tensorboard_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient, "update_tensorboard"
    ) as update_tensorboard_mock:
        update_tensorboard_lro_mock = mock.Mock(operation.Operation)
        update_tensorboard_lro_mock.result.return_value = gca_tensorboard.Tensorboard(
            name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME_UPDATE,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )
        update_tensorboard_mock.return_value = update_tensorboard_lro_mock
        yield update_tensorboard_mock


@pytest.fixture
def delete_tensorboard_mock():
    with mock.patch.object(
        tensorboard_service_client.TensorboardServiceClient, "delete_tensorboard"
    ) as delete_tensorboard_mock:
        delete_tensorboard_lro_mock = mock.Mock(operation.Operation)
        delete_tensorboard_lro_mock.result.return_value = (
            gca_tensorboard_service.DeleteTensorboardRequest(
                name=_TEST_NAME,
            )
        )
        delete_tensorboard_mock.return_value = delete_tensorboard_lro_mock
        yield delete_tensorboard_mock


_TEST_TENSORBOARD_EXPERIMENT = gca_tensorboard_experiment.TensorboardExperiment(
    name=_TEST_TENSORBOARD_EXPERIMENT_NAME,
    display_name=_TEST_DISPLAY_NAME,
)


@pytest.fixture
def get_tensorboard_experiment_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "get_tensorboard_experiment",
    ) as get_tensorboard_experiment__mock:
        get_tensorboard_experiment__mock.return_value = _TEST_TENSORBOARD_EXPERIMENT
        yield get_tensorboard_experiment__mock


@pytest.fixture
def create_tensorboard_experiment_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "create_tensorboard_experiment",
    ) as create_tensorboard_experiment_mock:
        create_tensorboard_experiment_mock.return_value = _TEST_TENSORBOARD_EXPERIMENT
        yield create_tensorboard_experiment_mock


@pytest.fixture
def delete_tensorboard_experiment_mock():
    with mock.patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "delete_tensorboard_experiment",
    ) as delete_tensorboard_experiment_mock:
        delete_tensorboard_lro_experiment_mock = mock.Mock(operation.Operation)
        delete_tensorboard_lro_experiment_mock.result.return_value = (
            gca_tensorboard_service.DeleteTensorboardExperimentRequest(
                name=_TEST_TENSORBOARD_EXPERIMENT_NAME,
            )
        )
        delete_tensorboard_experiment_mock.return_value = (
            delete_tensorboard_lro_experiment_mock
        )
        yield delete_tensorboard_experiment_mock


@pytest.fixture
def list_tensorboard_experiment_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "list_tensorboard_experiments",
    ) as list_tensorboard_experiment_mock:
        list_tensorboard_experiment_mock.return_value = [_TEST_TENSORBOARD_EXPERIMENT]
        yield list_tensorboard_experiment_mock


_TEST_TENSORBOARD_RUN = gca_tensorboard_run.TensorboardRun(
    name=_TEST_TENSORBOARD_RUN_NAME,
    display_name=_TEST_DISPLAY_NAME,
)


@pytest.fixture
def get_tensorboard_run_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "get_tensorboard_run",
    ) as get_tensorboard_run_mock:
        get_tensorboard_run_mock.return_value = _TEST_TENSORBOARD_RUN
        yield get_tensorboard_run_mock


@pytest.fixture
def create_tensorboard_run_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "create_tensorboard_run",
    ) as create_tensorboard_run_mock:
        create_tensorboard_run_mock.return_value = _TEST_TENSORBOARD_RUN
        yield create_tensorboard_run_mock


@pytest.fixture
def delete_tensorboard_run_mock():
    with mock.patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "delete_tensorboard_run",
    ) as delete_tensorboard_run_mock:
        delete_tensorboard_lro_run_mock = mock.Mock(operation.Operation)
        delete_tensorboard_lro_run_mock.result.return_value = (
            gca_tensorboard_service.DeleteTensorboardRunRequest(
                name=_TEST_TENSORBOARD_RUN_NAME,
            )
        )
        delete_tensorboard_run_mock.return_value = delete_tensorboard_lro_run_mock
        yield delete_tensorboard_run_mock


@pytest.fixture
def list_tensorboard_run_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "list_tensorboard_runs",
    ) as list_tensorboard_run_mock:
        list_tensorboard_run_mock.return_value = [
            gca_tensorboard_run.TensorboardRun(
                name=_TEST_TENSORBOARD_RUN_NAME,
                display_name=_TEST_DISPLAY_NAME,
            )
        ]
        yield list_tensorboard_run_mock


@pytest.fixture
def write_tensorboard_run_data_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "write_tensorboard_run_data",
    ) as write_tensorboard_run_data_mock:
        yield write_tensorboard_run_data_mock


_TEST_TENSORBOARD_TIME_SERIES = gca_tensorboard_time_series.TensorboardTimeSeries(
    name=_TEST_TENSORBOARD_TIME_SERIES_NAME,
    display_name=_TEST_TIME_SERIES_DISPLAY_NAME,
    value_type=gca_tensorboard_time_series.TensorboardTimeSeries.ValueType.SCALAR,
)


@pytest.fixture
def get_tensorboard_time_series_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "get_tensorboard_time_series",
    ) as get_tensorboard_time_series_mock:
        get_tensorboard_time_series_mock.return_value = _TEST_TENSORBOARD_TIME_SERIES
        yield get_tensorboard_time_series_mock


@pytest.fixture
def create_tensorboard_time_series_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "create_tensorboard_time_series",
    ) as create_tensorboard_time_series_mock:
        create_tensorboard_time_series_mock.return_value = _TEST_TENSORBOARD_TIME_SERIES
        yield create_tensorboard_time_series_mock


@pytest.fixture
def delete_tensorboard_time_series_mock():
    with mock.patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "delete_tensorboard_time_series",
    ) as delete_tensorboard_time_series_mock:
        delete_tensorboard_lro_time_series_mock = mock.Mock(operation.Operation)
        delete_tensorboard_lro_time_series_mock.result.return_value = (
            gca_tensorboard_service.DeleteTensorboardTimeSeriesRequest(
                name=_TEST_TENSORBOARD_TIME_SERIES_NAME,
            )
        )
        delete_tensorboard_time_series_mock.return_value = (
            delete_tensorboard_lro_time_series_mock
        )
        yield delete_tensorboard_time_series_mock


@pytest.fixture
def list_tensorboard_time_series_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "list_tensorboard_time_series",
    ) as list_tensorboard_time_series_mock:
        list_tensorboard_time_series_mock.return_value = [_TEST_TENSORBOARD_TIME_SERIES]
        yield list_tensorboard_time_series_mock


_TEST_TENSORBOARD_TIME_SERIES_DATA = gca_tensorboard_data.TimeSeriesData(
    tensorboard_time_series_id=_TEST_TENSORBOARD_TIME_SERIES_ID,
    value_type=gca_tensorboard_time_series.TensorboardTimeSeries.ValueType.SCALAR,
    values=[
        gca_tensorboard_data.TimeSeriesDataPoint(
            scalar=gca_tensorboard_data.Scalar(value=1.0),
            step=1,
            wall_time=utils.get_timestamp_proto(),
        )
    ],
)


@pytest.fixture
def batch_read_tensorboard_time_series_mock():
    with patch.object(
        tensorboard_service_client.TensorboardServiceClient,
        "batch_read_tensorboard_time_series_data",
    ) as batch_read_tensorboard_time_series_data_mock:
        batch_read_tensorboard_time_series_data_mock.return_value = (
            gca_tensorboard_service.BatchReadTensorboardTimeSeriesDataResponse(
                time_series_data=[_TEST_TENSORBOARD_TIME_SERIES_DATA]
            )
        )
        yield batch_read_tensorboard_time_series_data_mock


@pytest.mark.usefixtures("google_auth_mock")
class TestTensorboard:
    def setup_method(self):
        reload(initializer)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    def test_init_tensorboard(self, get_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.Tensorboard(tensorboard_name=_TEST_NAME)
        get_tensorboard_mock.assert_called_once_with(
            name=_TEST_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_with_id_only_with_project_and_location(
        self, get_tensorboard_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.Tensorboard(
            tensorboard_name=_TEST_ID, project=_TEST_PROJECT, location=_TEST_LOCATION
        )
        get_tensorboard_mock.assert_called_once_with(
            name=_TEST_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_with_project_and_location(self, get_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.Tensorboard(
            tensorboard_name=_TEST_NAME, project=_TEST_PROJECT, location=_TEST_LOCATION
        )
        get_tensorboard_mock.assert_called_once_with(
            name=_TEST_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_with_alt_project_and_location(self, get_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.Tensorboard(
            tensorboard_name=_TEST_NAME,
            project=_TEST_ALT_PROJECT,
            location=_TEST_LOCATION,
        )
        get_tensorboard_mock.assert_called_once_with(
            name=_TEST_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_with_alt_location(self, get_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT, location=_TEST_ALT_LOCATION)
        tensorboard.Tensorboard(
            tensorboard_name=_TEST_NAME,
        )
        get_tensorboard_mock.assert_called_once_with(
            name=_TEST_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_with_project_and_alt_location(self):
        aiplatform.init(project=_TEST_PROJECT)
        with pytest.raises(RuntimeError):
            tensorboard.Tensorboard(
                tensorboard_name=_TEST_NAME,
                project=_TEST_PROJECT,
                location=_TEST_ALT_LOCATION,
            )

    def test_init_tensorboard_with_location_override(self, get_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT, location=_TEST_LOCATION)
        tensorboard.Tensorboard(tensorboard_name=_TEST_ID, location=_TEST_ALT_LOCATION)
        get_tensorboard_mock.assert_called_once_with(
            name=_TEST_ALT_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_init_tensorboard_with_invalid_name(self):
        with pytest.raises(ValueError):
            aiplatform.init(project=_TEST_PROJECT, location=_TEST_LOCATION)
            tensorboard.Tensorboard(tensorboard_name=_TEST_INVALID_NAME)

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_create_tensorboard_with_default_encryption_key(
        self, create_tensorboard_mock
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            encryption_spec_key_name=_TEST_ENCRYPTION_KEY_NAME,
        )

        tensorboard.Tensorboard.create(
            display_name=_TEST_DISPLAY_NAME,
            create_request_timeout=None,
        )

        expected_tensorboard = gca_tensorboard.Tensorboard(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )

        create_tensorboard_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            tensorboard=expected_tensorboard,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_create_tensorboard(self, create_tensorboard_mock):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.Tensorboard.create(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec_key_name=_TEST_ENCRYPTION_KEY_NAME,
            create_request_timeout=None,
        )

        expected_tensorboard = gca_tensorboard.Tensorboard(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )

        create_tensorboard_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            tensorboard=expected_tensorboard,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_create_tensorboard_with_timeout(self, create_tensorboard_mock):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.Tensorboard.create(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec_key_name=_TEST_ENCRYPTION_KEY_NAME,
            create_request_timeout=180.0,
        )

        expected_tensorboard = gca_tensorboard.Tensorboard(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )

        create_tensorboard_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            tensorboard=expected_tensorboard,
            metadata=_TEST_REQUEST_METADATA,
            timeout=180.0,
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_create_tensorboard_with_timeout_not_explicitly_set(
        self, create_tensorboard_mock
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.Tensorboard.create(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec_key_name=_TEST_ENCRYPTION_KEY_NAME,
        )

        expected_tensorboard = gca_tensorboard.Tensorboard(
            display_name=_TEST_DISPLAY_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )

        create_tensorboard_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            tensorboard=expected_tensorboard,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_delete_tensorboard(self, delete_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT)

        my_tensorboard = tensorboard.Tensorboard(tensorboard_name=_TEST_NAME)

        my_tensorboard.delete()

        delete_tensorboard_mock.assert_called_once_with(
            name=my_tensorboard.resource_name
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_update_tensorboard_display_name(self, update_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT)

        my_tensorboard = tensorboard.Tensorboard(tensorboard_name=_TEST_NAME)
        my_tensorboard.update(display_name=_TEST_DISPLAY_NAME_UPDATE)

        expected_tensorboard = gca_tensorboard.Tensorboard(
            name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME_UPDATE,
        )
        update_tensorboard_mock.assert_called_once_with(
            update_mask=field_mask_pb2.FieldMask(paths=["display_name"]),
            tensorboard=expected_tensorboard,
            metadata=_TEST_REQUEST_METADATA,
        )

    @pytest.mark.usefixtures("get_tensorboard_mock")
    def test_update_tensorboard_encryption_spec(self, update_tensorboard_mock):
        aiplatform.init(project=_TEST_PROJECT)

        my_tensorboard = tensorboard.Tensorboard(tensorboard_name=_TEST_NAME)
        my_tensorboard.update(encryption_spec_key_name=_TEST_ENCRYPTION_KEY_NAME)

        expected_tensorboard = gca_tensorboard.Tensorboard(
            name=_TEST_NAME,
            encryption_spec=_TEST_ENCRYPTION_SPEC,
        )
        update_tensorboard_mock.assert_called_once_with(
            update_mask=field_mask_pb2.FieldMask(paths=["encryption_spec"]),
            tensorboard=expected_tensorboard,
            metadata=_TEST_REQUEST_METADATA,
        )


@pytest.mark.usefixtures("google_auth_mock")
class TestTensorboardExperiment:
    def setup_method(self):
        reload(initializer)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    def test_init_tensorboard_experiment(self, get_tensorboard_experiment_mock):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardExperiment(
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_NAME
        )
        get_tensorboard_experiment_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_EXPERIMENT_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_experiment_with_tensorboard(
        self, get_tensorboard_experiment_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardExperiment(
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_id=_TEST_ID,
        )
        get_tensorboard_experiment_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_EXPERIMENT_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_experiment_with_id_only_with_project_and_location(
        self, get_tensorboard_experiment_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardExperiment(
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_id=_TEST_ID,
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
        )
        get_tensorboard_experiment_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_EXPERIMENT_NAME, retry=base._DEFAULT_RETRY
        )

    def test_create_tensorboard_experiment(
        self, create_tensorboard_experiment_mock, get_tensorboard_experiment_mock
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardExperiment.create(
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME,
            create_request_timeout=None,
        )

        expected_tensorboard_experiment = (
            gca_tensorboard_experiment.TensorboardExperiment(
                display_name=_TEST_DISPLAY_NAME,
            )
        )

        create_tensorboard_experiment_mock.assert_called_once_with(
            parent=_TEST_NAME,
            tensorboard_experiment=expected_tensorboard_experiment,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

        get_tensorboard_experiment_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_EXPERIMENT_NAME, retry=base._DEFAULT_RETRY
        )

    def test_create_tensorboard_experiment_with_timeout(
        self, create_tensorboard_experiment_mock, get_tensorboard_experiment_mock
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardExperiment.create(
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME,
            create_request_timeout=180.0,
        )

        expected_tensorboard_experiment = (
            gca_tensorboard_experiment.TensorboardExperiment(
                display_name=_TEST_DISPLAY_NAME,
            )
        )

        create_tensorboard_experiment_mock.assert_called_once_with(
            parent=_TEST_NAME,
            tensorboard_experiment=expected_tensorboard_experiment,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            metadata=_TEST_REQUEST_METADATA,
            timeout=180.0,
        )

    def test_create_tensorboard_experiment_with_timeout_not_explicitly_set(
        self, create_tensorboard_experiment_mock, get_tensorboard_experiment_mock
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardExperiment.create(
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_name=_TEST_NAME,
            display_name=_TEST_DISPLAY_NAME,
        )

        expected_tensorboard_experiment = (
            gca_tensorboard_experiment.TensorboardExperiment(
                display_name=_TEST_DISPLAY_NAME,
            )
        )

        create_tensorboard_experiment_mock.assert_called_once_with(
            parent=_TEST_NAME,
            tensorboard_experiment=expected_tensorboard_experiment,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

    @pytest.mark.usefixtures("get_tensorboard_experiment_mock")
    def test_delete_tensorboard_experiement(self, delete_tensorboard_experiment_mock):
        aiplatform.init(project=_TEST_PROJECT)

        my_tensorboard_experiment = tensorboard.TensorboardExperiment(
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_NAME
        )

        my_tensorboard_experiment.delete()

        delete_tensorboard_experiment_mock.assert_called_once_with(
            name=my_tensorboard_experiment.resource_name
        )

    def test_list_tensorboard_experiments(self, list_tensorboard_experiment_mock):
        aiplatform.init(project=_TEST_PROJECT)

        tensorboard.TensorboardExperiment.list(tensorboard_name=_TEST_NAME)

        list_tensorboard_experiment_mock.assert_called_once_with(
            request={"parent": _TEST_NAME}
        )


@pytest.mark.usefixtures("google_auth_mock")
class TestTensorboardRun:
    def setup_method(self):
        reload(initializer)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    @pytest.mark.usefixtures("list_tensorboard_time_series_mock")
    def test_init_tensorboard_run(self, get_tensorboard_run_mock):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardRun(tensorboard_run_name=_TEST_TENSORBOARD_RUN_NAME)
        get_tensorboard_run_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_RUN_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures("list_tensorboard_time_series_mock")
    def test_init_tensorboard_run_with_tensorboard_and_experiment(
        self, get_tensorboard_run_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardRun(
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_id=_TEST_ID,
        )
        get_tensorboard_run_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_RUN_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_run_with_id_only_with_project_and_location(
        self, get_tensorboard_run_mock, list_tensorboard_time_series_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardRun(
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_id=_TEST_ID,
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
        )
        get_tensorboard_run_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_RUN_NAME, retry=base._DEFAULT_RETRY
        )
        list_tensorboard_time_series_mock.assert_called_once_with(
            request={"parent": _TEST_TENSORBOARD_RUN_NAME}
        )

    @pytest.mark.usefixtures("list_tensorboard_time_series_mock")
    def test_create_tensorboard_run(
        self,
        create_tensorboard_run_mock,
        get_tensorboard_run_mock,
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardRun.create(
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_NAME,
            create_request_timeout=None,
        )

        expected_tensorboard_run = gca_tensorboard_run.TensorboardRun(
            display_name=_TEST_TENSORBOARD_RUN_ID,
        )

        create_tensorboard_run_mock.assert_called_once_with(
            parent=_TEST_TENSORBOARD_EXPERIMENT_NAME,
            tensorboard_run=expected_tensorboard_run,
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

        get_tensorboard_run_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_RUN_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures("list_tensorboard_time_series_mock")
    def test_create_tensorboard_run_with_timeout(
        self, create_tensorboard_run_mock, get_tensorboard_run_mock
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardRun.create(
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_NAME,
            create_request_timeout=180.0,
        )

        expected_tensorboard_run = gca_tensorboard_run.TensorboardRun(
            display_name=_TEST_TENSORBOARD_RUN_ID,
        )

        create_tensorboard_run_mock.assert_called_once_with(
            parent=_TEST_TENSORBOARD_EXPERIMENT_NAME,
            tensorboard_run=expected_tensorboard_run,
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            metadata=_TEST_REQUEST_METADATA,
            timeout=180.0,
        )

    @pytest.mark.usefixtures("list_tensorboard_time_series_mock")
    def test_create_tensorboard_run_with_timeout_not_explicitly_set(
        self, create_tensorboard_run_mock, get_tensorboard_run_mock
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardRun.create(
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_NAME,
        )

        expected_tensorboard_run = gca_tensorboard_run.TensorboardRun(
            display_name=_TEST_TENSORBOARD_RUN_ID,
        )

        create_tensorboard_run_mock.assert_called_once_with(
            parent=_TEST_TENSORBOARD_EXPERIMENT_NAME,
            tensorboard_run=expected_tensorboard_run,
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            metadata=_TEST_REQUEST_METADATA,
            timeout=None,
        )

    @pytest.mark.usefixtures(
        "get_tensorboard_run_mock", "list_tensorboard_time_series_mock"
    )
    def test_delete_tensorboard_run(self, delete_tensorboard_run_mock):
        aiplatform.init(project=_TEST_PROJECT)

        my_tensorboard_run = tensorboard.TensorboardRun(
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_NAME
        )

        my_tensorboard_run.delete()

        delete_tensorboard_run_mock.assert_called_once_with(
            name=my_tensorboard_run.resource_name
        )

    def test_list_tensorboard_runs(
        self, list_tensorboard_run_mock, list_tensorboard_time_series_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)

        tensorboard.TensorboardRun.list(
            tensorboard_experiment_name=_TEST_TENSORBOARD_EXPERIMENT_NAME
        )

        list_tensorboard_run_mock.assert_called_once_with(
            request={"parent": _TEST_TENSORBOARD_EXPERIMENT_NAME}
        )

        list_tensorboard_time_series_mock.assert_called_once_with(
            request={"parent": _TEST_TENSORBOARD_RUN_NAME}
        )

    @pytest.mark.usefixtures(
        "get_tensorboard_run_mock", "list_tensorboard_time_series_mock"
    )
    def test_write_tensorboard_run_data(self, write_tensorboard_run_data_mock):
        aiplatform.init(project=_TEST_PROJECT)

        tb_run = tensorboard.TensorboardRun(
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_NAME
        )

        timestamp = utils.get_timestamp_proto()
        tb_run.write_tensorboard_scalar_data(
            time_series_data={"accuracy": 0.9}, step=1, wall_time=timestamp
        )

        expected_time_series_data = [
            gca_tensorboard_data.TimeSeriesData(
                tensorboard_time_series_id=_TEST_TENSORBOARD_TIME_SERIES_ID,
                value_type=gca_tensorboard_time_series.TensorboardTimeSeries.ValueType.SCALAR,
                values=[
                    gca_tensorboard_data.TimeSeriesDataPoint(
                        scalar=gca_tensorboard_data.Scalar(value=0.9),
                        wall_time=timestamp,
                        step=1,
                    )
                ],
            ),
        ]

        write_tensorboard_run_data_mock.assert_called_once_with(
            tensorboard_run=_TEST_TENSORBOARD_RUN_NAME,
            time_series_data=expected_time_series_data,
        )

    @pytest.mark.usefixtures(
        "get_tensorboard_run_mock", "list_tensorboard_time_series_mock"
    )
    def test_read_tensorboard_time_series(
        self, batch_read_tensorboard_time_series_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)

        tb_run = tensorboard.TensorboardRun(
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_NAME
        )

        ts_data = tb_run.read_time_series_data()

        true_ts_data = {
            _TEST_TIME_SERIES_DISPLAY_NAME: _TEST_TENSORBOARD_TIME_SERIES_DATA
        }

        batch_read_tensorboard_time_series_mock.assert_called_once_with(
            request=gca_tensorboard_service.BatchReadTensorboardTimeSeriesDataRequest(
                tensorboard=_TEST_NAME,
                time_series=[_TEST_TENSORBOARD_TIME_SERIES_NAME],
            )
        )

        assert ts_data == true_ts_data


@pytest.mark.usefixtures("google_auth_mock")
class TestTensorboardTimeSeries:
    def setup_method(self):
        reload(initializer)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    def test_init_tensorboard_time_series(self, get_tensorboard_time_series_mock):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardTimeSeries(
            tensorboard_time_series_name=_TEST_TENSORBOARD_TIME_SERIES_NAME
        )
        get_tensorboard_time_series_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_TIME_SERIES_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_time_series_with_tensorboard_and_experiment_and_run(
        self, get_tensorboard_time_series_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardTimeSeries(
            tensorboard_time_series_name=_TEST_TENSORBOARD_TIME_SERIES_ID,
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_id=_TEST_ID,
        )
        get_tensorboard_time_series_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_TIME_SERIES_NAME, retry=base._DEFAULT_RETRY
        )

    def test_init_tensorboard_time_series_with_id_only_with_project_and_location(
        self, get_tensorboard_time_series_mock
    ):
        aiplatform.init(project=_TEST_PROJECT)
        tensorboard.TensorboardTimeSeries(
            tensorboard_time_series_name=_TEST_TENSORBOARD_TIME_SERIES_ID,
            tensorboard_run_id=_TEST_TENSORBOARD_RUN_ID,
            tensorboard_experiment_id=_TEST_TENSORBOARD_EXPERIMENT_ID,
            tensorboard_id=_TEST_ID,
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
        )
        get_tensorboard_time_series_mock.assert_called_once_with(
            name=_TEST_TENSORBOARD_TIME_SERIES_NAME, retry=base._DEFAULT_RETRY
        )

    def test_create_tensorboard_time_series(
        self,
        create_tensorboard_time_series_mock,
    ):

        aiplatform.init(
            project=_TEST_PROJECT,
        )

        tensorboard.TensorboardTimeSeries.create(
            display_name=_TEST_TIME_SERIES_DISPLAY_NAME,
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_NAME,
        )

        expected_tensorboard_time_series = gca_tensorboard_time_series.TensorboardTimeSeries(
            display_name=_TEST_TIME_SERIES_DISPLAY_NAME,
            value_type=gca_tensorboard_time_series.TensorboardTimeSeries.ValueType.SCALAR,
            plugin_name="scalars",
        )

        create_tensorboard_time_series_mock.assert_called_once_with(
            parent=_TEST_TENSORBOARD_RUN_NAME,
            tensorboard_time_series=expected_tensorboard_time_series,
        )

    @pytest.mark.usefixtures("get_tensorboard_time_series_mock")
    def test_delete_tensorboard_time_series(self, delete_tensorboard_time_series_mock):
        aiplatform.init(project=_TEST_PROJECT)

        my_tensorboard_time_series = tensorboard.TensorboardTimeSeries(
            tensorboard_time_series_name=_TEST_TENSORBOARD_TIME_SERIES_NAME
        )

        my_tensorboard_time_series.delete()

        delete_tensorboard_time_series_mock.assert_called_once_with(
            name=my_tensorboard_time_series.resource_name
        )

    def test_list_tensorboard_time_series(self, list_tensorboard_time_series_mock):
        aiplatform.init(project=_TEST_PROJECT)

        tensorboard.TensorboardTimeSeries.list(
            tensorboard_run_name=_TEST_TENSORBOARD_RUN_NAME
        )

        list_tensorboard_time_series_mock.assert_called_once_with(
            request={"parent": _TEST_TENSORBOARD_RUN_NAME}
        )



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1.schema.predict.prediction",
    manifest={
        "ClassificationPredictionResult",
    },
)


class ClassificationPredictionResult(proto.Message):
    r"""Prediction output format for Image and Text Classification.

    Attributes:
        ids (MutableSequence[int]):
            The resource IDs of the AnnotationSpecs that
            had been identified.
        display_names (MutableSequence[str]):
            The display names of the AnnotationSpecs that
            had been identified, order matches the IDs.
        confidences (MutableSequence[float]):
            The Model's confidences in correctness of the
            predicted IDs, higher value means higher
            confidence. Order matches the Ids.
    """

    ids: MutableSequence[int] = proto.RepeatedField(
        proto.INT64,
        number=1,
    )
    display_names: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=2,
    )
    confidences: MutableSequence[float] = proto.RepeatedField(
        proto.FLOAT,
        number=3,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
import os
import pickle
from importlib import reload
from unittest import mock
from unittest.mock import patch
import uuid

from google.auth import credentials as auth_credentials
from google.cloud import aiplatform
from google.cloud.aiplatform import base
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform.metadata import constants
from google.cloud.aiplatform.metadata import metadata
from google.cloud.aiplatform.metadata import _models
from google.cloud.aiplatform.models import Model
from google.cloud.aiplatform_v1 import Artifact as GapicArtifact
from google.cloud.aiplatform_v1 import MetadataStore as GapicMetadataStore
from google.cloud.aiplatform_v1 import MetadataServiceClient
import numpy as np
import pytest
import sklearn
from sklearn.datasets import make_classification
from sklearn.linear_model import LinearRegression
import tensorflow as tf
import xgboost as xgb


# project
_TEST_PROJECT = "test-project"
_TEST_LOCATION = "us-central1"
_TEST_BUCKET_NAME = "gs://test-bucket"
_TEST_PARENT = (
    f"projects/{_TEST_PROJECT}/locations/{_TEST_LOCATION}/metadataStores/default"
)
_TEST_CREDENTIALS = mock.Mock(spec=auth_credentials.AnonymousCredentials())


# artifact
_TEST_ARTIFACT_ID = "test-model-id"
_TEST_URI = "gs://test-uri"
_TEST_DISPLAY_NAME = "test-model-display-name"

_TEST_ARTIFACT_ID = "test-model-id"
_TEST_ARTIFACT_NAME = f"{_TEST_PARENT}/artifacts/{_TEST_ARTIFACT_ID}"

_TEST_TIMESTAMP = "2022-11-30-00-00-00"
_TEST_DATETIME = datetime.datetime.strptime(_TEST_TIMESTAMP, "%Y-%m-%d-%H-%M-%S")

_TEST_UUID = uuid.UUID("fa2db23f-1b13-412d-beea-94602448e4ce")

_TEST_INPUT_EXAMPLE = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])

_TEST_MODEL_NAME = (
    f"projects/{_TEST_PROJECT}/locations/{_TEST_LOCATION}/models/{_TEST_ARTIFACT_ID}"
)


@pytest.fixture
def mock_datetime_now(monkeypatch):
    class DateTime:
        @classmethod
        def now(cls):
            return _TEST_DATETIME

    monkeypatch.setattr(datetime, "datetime", DateTime)


@pytest.fixture
def mock_uuid():
    with patch.object(uuid, "uuid4", return_value=_TEST_UUID) as mock_uuid:
        yield mock_uuid


@pytest.fixture
def mock_keras_save_model():
    with patch.object(tf.keras.models.Sequential, "save") as mock_keras_save_model:
        yield mock_keras_save_model


@pytest.fixture
def mock_tf_save_model():
    with patch("tensorflow.saved_model.save") as mock_tf_save_model:
        yield mock_tf_save_model


@pytest.fixture
def mock_storage_blob_upload_from_filename():
    with patch(
        "google.cloud.storage.Blob.upload_from_filename"
    ) as mock_blob_upload_from_filename, patch(
        "google.cloud.storage.Bucket.exists", return_value=True
    ):
        yield mock_blob_upload_from_filename


@pytest.fixture
def mock_storage_blob_download_sklearn_model_file():
    def create_model_file(filename):
        train_x = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        train_y = np.dot(train_x, np.array([1, 2])) + 3
        model = LinearRegression()
        model.fit(train_x, train_y)
        with open(filename, "wb") as model_file:
            pickle.dump(model, model_file)

    with patch(
        "google.cloud.storage.Blob.download_to_filename", wraps=create_model_file
    ) as mock_blob_download_to_filename, patch(
        "google.cloud.storage.Bucket.exists", return_value=True
    ):
        yield mock_blob_download_to_filename


@pytest.fixture
def mock_storage_blob_download_xgboost_booster_file():
    def create_model_file(filename):
        x, y = make_classification()
        dtrain = xgb.DMatrix(data=x, label=y)
        booster = xgb.train(
            params={"num_parallel_tree": 4, "subsample": 0.5, "num_class": 2},
            dtrain=dtrain,
        )
        booster.save_model(filename)

    with patch(
        "google.cloud.storage.Blob.download_to_filename", wraps=create_model_file
    ) as mock_blob_download_to_filename, patch(
        "google.cloud.storage.Bucket.exists", return_value=True
    ):
        yield mock_blob_download_to_filename


@pytest.fixture
def mock_storage_blob_download_xgboost_xgbmodel_file():
    def create_model_file(filename):
        x, y = make_classification()
        model = xgb.XGBClassifier()
        model.fit(x, y)
        model.save_model(filename)

    with patch(
        "google.cloud.storage.Blob.download_to_filename", wraps=create_model_file
    ) as mock_blob_download_to_filename, patch(
        "google.cloud.storage.Bucket.exists", return_value=True
    ):
        yield mock_blob_download_to_filename


@pytest.fixture
def mock_storage_blob_download_input_example():
    def create_input_example_file(filename):
        filepath, _ = os.path.split(filename)
        _models._save_input_example(_TEST_INPUT_EXAMPLE, filepath)

    with patch(
        "google.cloud.storage.Blob.download_to_filename",
        wraps=create_input_example_file,
    ) as mock_blob_download_to_filename, patch(
        "google.cloud.storage.Bucket.exists", return_value=True
    ):
        yield mock_blob_download_to_filename


@pytest.fixture
def mock_load_tensorflow_keras_model():
    with patch("tensorflow.keras.models.load_model") as load_tensorflow_keras_model:
        yield load_tensorflow_keras_model


@pytest.fixture
def mock_load_tensorflow_module_model():
    with patch("tensorflow.saved_model.load") as load_tensorflow_keras_model:
        yield load_tensorflow_keras_model


_TEST_SKLEARN_MODEL_ARTIFACT = GapicArtifact(
    name=_TEST_ARTIFACT_NAME,
    uri=_TEST_URI,
    display_name=_TEST_DISPLAY_NAME,
    schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
    schema_version=constants._DEFAULT_SCHEMA_VERSION,
    state=GapicArtifact.State.LIVE,
    metadata={
        "frameworkName": "sklearn",
        "frameworkVersion": "1.0",
        "modelFile": "model.pkl",
        "modelClass": "sklearn.linear_model._base.LinearRegression",
    },
)


@pytest.fixture
def create_experiment_model_artifact_mock():
    with patch.object(MetadataServiceClient, "create_artifact") as create_artifact_mock:
        create_artifact_mock.return_value = _TEST_SKLEARN_MODEL_ARTIFACT
        yield create_artifact_mock


@pytest.fixture
def get_sklearn_model_artifact_mock():
    with patch.object(MetadataServiceClient, "get_artifact") as get_artifact_mock:
        get_artifact_mock.return_value = _TEST_SKLEARN_MODEL_ARTIFACT
        yield get_artifact_mock


_TEST_XGBOOST_BOOSTER_ARTIFACT = GapicArtifact(
    name=_TEST_ARTIFACT_NAME,
    uri=_TEST_URI,
    display_name=_TEST_DISPLAY_NAME,
    schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
    schema_version=constants._DEFAULT_SCHEMA_VERSION,
    state=GapicArtifact.State.LIVE,
    metadata={
        "frameworkName": "xgboost",
        "frameworkVersion": "1.5",
        "modelFile": "model.bst",
        "modelClass": "xgboost.core.Booster",
    },
)


@pytest.fixture
def get_xgboost_booster_artifact_mock():
    with patch.object(MetadataServiceClient, "get_artifact") as get_artifact_mock:
        get_artifact_mock.return_value = _TEST_XGBOOST_BOOSTER_ARTIFACT
        yield get_artifact_mock


_TEST_XGBOOST_XGBMODEL_ARTIFACT = GapicArtifact(
    name=_TEST_ARTIFACT_NAME,
    uri=_TEST_URI,
    display_name=_TEST_DISPLAY_NAME,
    schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
    schema_version=constants._DEFAULT_SCHEMA_VERSION,
    state=GapicArtifact.State.LIVE,
    metadata={
        "frameworkName": "xgboost",
        "frameworkVersion": "1.5",
        "modelFile": "model.bst",
        "modelClass": "xgboost.sklearn.XGBClassifier",
    },
)


@pytest.fixture
def get_xgboost_xgbmodel_artifact_mock():
    with patch.object(MetadataServiceClient, "get_artifact") as get_artifact_mock:
        get_artifact_mock.return_value = _TEST_XGBOOST_XGBMODEL_ARTIFACT
        yield get_artifact_mock


_TEST_TENSORFLOW_KERAS_ARTIFACT = GapicArtifact(
    name=_TEST_ARTIFACT_NAME,
    uri=_TEST_URI,
    display_name=_TEST_DISPLAY_NAME,
    schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
    schema_version=constants._DEFAULT_SCHEMA_VERSION,
    state=GapicArtifact.State.LIVE,
    metadata={
        "frameworkName": "tensorflow",
        "frameworkVersion": "2.8",
        "modelFile": "saved_model",
        "modelClass": "tensorflow.keras.Model",
    },
)


@pytest.fixture
def get_tensorflow_keras_artifact_mock():
    with patch.object(MetadataServiceClient, "get_artifact") as get_artifact_mock:
        get_artifact_mock.return_value = _TEST_TENSORFLOW_KERAS_ARTIFACT
        yield get_artifact_mock


_TEST_TENSORFLOW_MODULE_ARTIFACT = GapicArtifact(
    name=_TEST_ARTIFACT_NAME,
    uri=_TEST_URI,
    display_name=_TEST_DISPLAY_NAME,
    schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
    schema_version=constants._DEFAULT_SCHEMA_VERSION,
    state=GapicArtifact.State.LIVE,
    metadata={
        "frameworkName": "tensorflow",
        "frameworkVersion": "2.8",
        "modelFile": "saved_model",
        "modelClass": "tensorflow.Module",
    },
)


@pytest.fixture
def get_tensorflow_module_artifact_mock():
    with patch.object(MetadataServiceClient, "get_artifact") as get_artifact_mock:
        get_artifact_mock.return_value = _TEST_TENSORFLOW_MODULE_ARTIFACT
        yield get_artifact_mock


@pytest.fixture
def model_upload_mock():
    with patch.object(Model, "upload") as upload_model_mock:
        yield upload_model_mock


@pytest.fixture
def get_metadata_store_mock():
    with patch.object(
        MetadataServiceClient, "get_metadata_store"
    ) as get_metadata_store_mock:
        get_metadata_store_mock.return_value = GapicMetadataStore(name=_TEST_PARENT)
        yield get_metadata_store_mock


class TestModels:
    def setup_method(self):
        reload(initializer)
        reload(metadata)
        reload(aiplatform)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    @pytest.mark.usefixtures(
        "mock_datetime_now",
        "mock_uuid",
        "get_metadata_store_mock",
    )
    def test_save_model_sklearn(
        self,
        mock_storage_blob_upload_from_filename,
        create_experiment_model_artifact_mock,
        get_sklearn_model_artifact_mock,
    ):
        train_x = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        train_y = np.dot(train_x, np.array([1, 2])) + 3
        model = LinearRegression()
        model.fit(train_x, train_y)

        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        aiplatform.save_model(model, _TEST_ARTIFACT_ID)

        # Verify that the model file is correctly uploaded to gcs
        upload_file_path = mock_storage_blob_upload_from_filename.call_args[1][
            "filename"
        ]
        assert upload_file_path.endswith("model.pkl")

        # Verify the model artifact is created correctly
        expected_artifact = GapicArtifact(
            uri=f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}-sklearn-model",
            schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
            schema_version=constants._DEFAULT_SCHEMA_VERSION,
            metadata={
                "frameworkName": "sklearn",
                "frameworkVersion": sklearn.__version__,
                "modelFile": "model.pkl",
                "modelClass": "sklearn.linear_model._base.LinearRegression",
            },
            state=GapicArtifact.State.LIVE,
        )
        create_experiment_model_artifact_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            artifact=expected_artifact,
            artifact_id=_TEST_ARTIFACT_ID,
        )

        get_sklearn_model_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures(
        "mock_storage_blob_upload_from_filename",
        "get_sklearn_model_artifact_mock",
        "get_metadata_store_mock",
    )
    def test_save_model_with_all_args(
        self,
        create_experiment_model_artifact_mock,
    ):
        train_x = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        train_y = np.dot(train_x, np.array([1, 2])) + 3
        model = LinearRegression()
        model.fit(train_x, train_y)

        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        aiplatform.save_model(
            model=model,
            artifact_id=_TEST_ARTIFACT_ID,
            uri=_TEST_URI,
            display_name=_TEST_DISPLAY_NAME,
            input_example=_TEST_INPUT_EXAMPLE,
        )

        # Verify the model artifact is created correctly
        expected_artifact = GapicArtifact(
            display_name=_TEST_DISPLAY_NAME,
            uri=_TEST_URI,
            schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
            schema_version=constants._DEFAULT_SCHEMA_VERSION,
            metadata={
                "frameworkName": "sklearn",
                "frameworkVersion": sklearn.__version__,
                "modelFile": "model.pkl",
                "modelClass": "sklearn.linear_model._base.LinearRegression",
                "predictSchemata": {"instanceSchemaUri": f"{_TEST_URI}/instance.yaml"},
            },
            state=GapicArtifact.State.LIVE,
        )
        create_experiment_model_artifact_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            artifact=expected_artifact,
            artifact_id=_TEST_ARTIFACT_ID,
        )

    @pytest.mark.usefixtures(
        "mock_datetime_now",
        "mock_uuid",
        "get_metadata_store_mock",
    )
    def test_save_model_xgboost_booster(
        self,
        mock_storage_blob_upload_from_filename,
        create_experiment_model_artifact_mock,
        get_xgboost_booster_artifact_mock,
    ):
        # Fix the bug that xgb.__version__ in third_party returns a byte not string
        xgb.__version__ = "1.5.1"

        x, y = make_classification()
        dtrain = xgb.DMatrix(data=x, label=y)
        booster = xgb.train(
            params={"num_parallel_tree": 4, "subsample": 0.5, "num_class": 2},
            dtrain=dtrain,
        )

        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        aiplatform.save_model(booster, _TEST_ARTIFACT_ID)

        # Verify that the model file is correctly uploaded to gcs
        upload_file_path = mock_storage_blob_upload_from_filename.call_args[1][
            "filename"
        ]
        assert upload_file_path.endswith("model.bst")

        # Verify the model artifact is created correctly
        expected_artifact = GapicArtifact(
            uri=f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}-xgboost-model",
            schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
            schema_version=constants._DEFAULT_SCHEMA_VERSION,
            metadata={
                "frameworkName": "xgboost",
                "frameworkVersion": xgb.__version__,
                "modelFile": "model.bst",
                "modelClass": "xgboost.core.Booster",
            },
            state=GapicArtifact.State.LIVE,
        )
        create_experiment_model_artifact_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            artifact=expected_artifact,
            artifact_id=_TEST_ARTIFACT_ID,
        )

        get_xgboost_booster_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures(
        "mock_datetime_now",
        "mock_uuid",
        "get_metadata_store_mock",
    )
    def test_save_model_xgboost_xgbmodel(
        self,
        mock_storage_blob_upload_from_filename,
        create_experiment_model_artifact_mock,
        get_xgboost_xgbmodel_artifact_mock,
    ):
        # Fix the bug that xgb.__version__ in third_party returns a byte not string
        xgb.__version__ = "1.5.1"

        x, y = make_classification()
        xgb_model = xgb.XGBClassifier()
        xgb_model.fit(x, y)

        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        aiplatform.save_model(xgb_model, _TEST_ARTIFACT_ID)

        # Verify that the model file is correctly uploaded to gcs
        upload_file_path = mock_storage_blob_upload_from_filename.call_args[1][
            "filename"
        ]
        assert upload_file_path.endswith("model.bst")

        # Verify the model artifact is created correctly
        expected_artifact = GapicArtifact(
            uri=f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}-xgboost-model",
            schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
            schema_version=constants._DEFAULT_SCHEMA_VERSION,
            metadata={
                "frameworkName": "xgboost",
                "frameworkVersion": xgb.__version__,
                "modelFile": "model.bst",
                "modelClass": "xgboost.sklearn.XGBClassifier",
            },
            state=GapicArtifact.State.LIVE,
        )
        create_experiment_model_artifact_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            artifact=expected_artifact,
            artifact_id=_TEST_ARTIFACT_ID,
        )

        get_xgboost_xgbmodel_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures(
        "mock_datetime_now",
        "mock_uuid",
        "get_metadata_store_mock",
    )
    def test_save_model_tensorflow_keras(
        self,
        mock_keras_save_model,
        create_experiment_model_artifact_mock,
        get_tensorflow_keras_artifact_mock,
    ):
        x = np.random.random((100, 3))
        y = np.random.random((100, 1))
        model = tf.keras.Sequential(
            [tf.keras.layers.Dense(5, input_shape=(3,)), tf.keras.layers.Softmax()]
        )
        model.compile(optimizer="adam", loss="mean_squared_error")
        model.fit(x, y)

        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        aiplatform.save_model(model, _TEST_ARTIFACT_ID)

        mock_keras_save_model.assert_called_once_with(
            f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}"
            + "-tensorflow-model/saved_model",
        )

        # Verify the model artifact is created correctly
        expected_artifact = GapicArtifact(
            uri=f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}-tensorflow-model",
            schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
            schema_version=constants._DEFAULT_SCHEMA_VERSION,
            metadata={
                "frameworkName": "tensorflow",
                "frameworkVersion": tf.__version__,
                "modelFile": "saved_model",
                "modelClass": "tensorflow.keras.Model",
            },
            state=GapicArtifact.State.LIVE,
        )
        create_experiment_model_artifact_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            artifact=expected_artifact,
            artifact_id=_TEST_ARTIFACT_ID,
        )

        get_tensorflow_keras_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

    @pytest.mark.usefixtures(
        "mock_datetime_now",
        "mock_uuid",
        "get_metadata_store_mock",
    )
    def test_save_model_tensorflow_module(
        self,
        mock_tf_save_model,
        create_experiment_model_artifact_mock,
        get_tensorflow_module_artifact_mock,
    ):
        class Adder(tf.Module):
            @tf.function(input_signature=[tf.TensorSpec(shape=[], dtype=tf.float32)])
            def add(self, x):
                return x + x

        model = Adder()

        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        aiplatform.save_model(model, _TEST_ARTIFACT_ID)

        mock_tf_save_model.assert_called_once_with(
            model,
            f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}"
            + "-tensorflow-model/saved_model",
        )

        # Verify the model artifact is created correctly
        expected_artifact = GapicArtifact(
            uri=f"{_TEST_BUCKET_NAME}/{_TEST_TIMESTAMP}-{_TEST_UUID.hex[:5]}-tensorflow-model",
            schema_title=constants.GOOGLE_EXPERIMENT_MODEL,
            schema_version=constants._DEFAULT_SCHEMA_VERSION,
            metadata={
                "frameworkName": "tensorflow",
                "frameworkVersion": tf.__version__,
                "modelFile": "saved_model",
                "modelClass": "tensorflow.Module",
            },
            state=GapicArtifact.State.LIVE,
        )
        create_experiment_model_artifact_mock.assert_called_once_with(
            parent=_TEST_PARENT,
            artifact=expected_artifact,
            artifact_id=_TEST_ARTIFACT_ID,
        )

        get_tensorflow_module_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

    def test_load_model_sklearn(
        self,
        mock_storage_blob_download_sklearn_model_file,
        get_sklearn_model_artifact_mock,
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        model = experiment_model.load_model()

        # Verify that the correct model artifact is retrieved by its ID
        get_sklearn_model_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

        # Verify that the model file is downloaded correctly
        download_file_path = mock_storage_blob_download_sklearn_model_file.call_args[1][
            "filename"
        ]
        assert download_file_path.endswith("model.pkl")

        # Verify the loaded model
        assert model.__class__.__name__ == "LinearRegression"

    def test_load_model_xgboost_booster(
        self,
        mock_storage_blob_download_xgboost_booster_file,
        get_xgboost_booster_artifact_mock,
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        # Fix the bug that xgb.__version__ in third_party returns a byte not string
        xgb.__version__ = "1.5.1"

        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        model = experiment_model.load_model()

        # Verify that the correct model artifact is retrieved by its ID
        get_xgboost_booster_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

        # Verify that the model file is downloaded correctly
        download_file_path = mock_storage_blob_download_xgboost_booster_file.call_args[
            1
        ]["filename"]
        assert download_file_path.endswith("model.bst")

        # Verify the loaded model
        assert model.__class__.__name__ == "Booster"

    def test_load_model_xgboost_xgbmodel(
        self,
        mock_storage_blob_download_xgboost_xgbmodel_file,
        get_xgboost_xgbmodel_artifact_mock,
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        # Fix the bug that xgb.__version__ in third_party returns a byte not string
        xgb.__version__ = "1.5.1"

        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        model = experiment_model.load_model()

        # Verify that the correct model artifact is retrieved by its ID
        get_xgboost_xgbmodel_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

        # Verify that the model file is downloaded correctly
        download_file_path = mock_storage_blob_download_xgboost_xgbmodel_file.call_args[
            1
        ]["filename"]
        assert download_file_path.endswith("model.bst")

        # Verify the loaded model
        assert model.__class__.__name__ == "XGBClassifier"

    def test_load_model_tensorflow_keras(
        self,
        mock_load_tensorflow_keras_model,
        get_tensorflow_keras_artifact_mock,
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        experiment_model.load_model()

        # Verify that the correct model artifact is retrieved by its ID
        get_tensorflow_keras_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

        # Verify that the model file is loaded correctly
        mock_load_tensorflow_keras_model.assert_called_once_with(
            f"{_TEST_URI}/saved_model",
        )

    def test_load_model_tensorflow_module(
        self,
        mock_load_tensorflow_module_model,
        get_tensorflow_module_artifact_mock,
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )
        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        experiment_model.load_model()

        # Verify that the correct model artifact is retrieved by its ID
        get_tensorflow_module_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )

        # Verify that the model file is loaded correctly
        mock_load_tensorflow_module_model.assert_called_once_with(
            f"{_TEST_URI}/saved_model",
        )

    def test_register_model_sklearn(
        self, model_upload_mock, get_sklearn_model_artifact_mock
    ):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )

        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        experiment_model.register_model(display_name=_TEST_DISPLAY_NAME)

        # Verify that the correct model artifact is retrieved by its ID
        get_sklearn_model_artifact_mock.assert_called_once_with(
            name=_TEST_ARTIFACT_NAME, retry=base._DEFAULT_RETRY
        )
        # register_model API calls Model.upload internally to register the model
        # Since Model.upload is tested in "test_models.py", here we only need to
        # make sure register_model is sending the right args to Model.upload
        model_upload_mock.assert_called_once_with(
            serving_container_image_uri="us-docker.pkg.dev/vertex-ai/prediction/sklearn-cpu.1-0:latest",
            artifact_uri=_TEST_URI,
            model_id=None,
            parent_model=None,
            is_default_version=True,
            version_aliases=None,
            version_description=None,
            display_name=_TEST_DISPLAY_NAME,
            description=None,
            labels=None,
            serving_container_predict_route=None,
            serving_container_health_route=None,
            serving_container_command=None,
            serving_container_args=None,
            serving_container_environment_variables=None,
            serving_container_ports=None,
            instance_schema_uri=None,
            parameters_schema_uri=None,
            prediction_schema_uri=None,
            explanation_metadata=None,
            explanation_parameters=None,
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            credentials=_TEST_CREDENTIALS,
            encryption_spec_key_name=None,
            staging_bucket=None,
            sync=True,
            upload_request_timeout=None,
        )

    @pytest.mark.usefixtures(
        "mock_storage_blob_download_input_example",
        "get_sklearn_model_artifact_mock",
    )
    def test_get_experiment_model_info(self):
        aiplatform.init(
            project=_TEST_PROJECT,
            location=_TEST_LOCATION,
            staging_bucket=_TEST_BUCKET_NAME,
            credentials=_TEST_CREDENTIALS,
        )
        experiment_model = aiplatform.get_experiment_model(_TEST_ARTIFACT_ID)
        model_info = experiment_model.get_model_info()

        expected_model_info = {
            "model_class": "sklearn.linear_model._base.LinearRegression",
            "framework_name": "sklearn",
            "framework_version": "1.0",
            "input_example": {
                "type": "numpy.ndarray",
                "data": _TEST_INPUT_EXAMPLE.tolist(),
            },
        }
        assert model_info == expected_model_info



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import abc
import datetime
import pathlib
import logging
import re
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Tuple
import uuid

from google.protobuf import timestamp_pb2

from google.api_core import client_options
from google.api_core import gapic_v1
from google.auth import credentials as auth_credentials
from google.cloud import storage

from google.cloud.aiplatform import compat
from google.cloud.aiplatform.constants import base as constants
from google.cloud.aiplatform import initializer

from google.cloud.aiplatform.compat.services import (
    dataset_service_client_v1beta1,
    endpoint_service_client_v1beta1,
    featurestore_online_serving_service_client_v1beta1,
    featurestore_service_client_v1beta1,
    index_service_client_v1beta1,
    index_endpoint_service_client_v1beta1,
    job_service_client_v1beta1,
    metadata_service_client_v1beta1,
    model_service_client_v1beta1,
    pipeline_service_client_v1beta1,
    prediction_service_client_v1beta1,
    tensorboard_service_client_v1beta1,
    vizier_service_client_v1beta1,
)
from google.cloud.aiplatform.compat.services import (
    dataset_service_client_v1,
    endpoint_service_client_v1,
    featurestore_online_serving_service_client_v1,
    featurestore_service_client_v1,
    index_service_client_v1,
    index_endpoint_service_client_v1,
    job_service_client_v1,
    metadata_service_client_v1,
    model_service_client_v1,
    pipeline_service_client_v1,
    prediction_service_client_v1,
    tensorboard_service_client_v1,
    vizier_service_client_v1,
)

from google.cloud.aiplatform.compat.types import (
    accelerator_type as gca_accelerator_type,
)

VertexAiServiceClient = TypeVar(
    "VertexAiServiceClient",
    # v1beta1
    dataset_service_client_v1beta1.DatasetServiceClient,
    endpoint_service_client_v1beta1.EndpointServiceClient,
    featurestore_online_serving_service_client_v1beta1.FeaturestoreOnlineServingServiceClient,
    featurestore_service_client_v1beta1.FeaturestoreServiceClient,
    index_service_client_v1beta1.IndexServiceClient,
    index_endpoint_service_client_v1beta1.IndexEndpointServiceClient,
    model_service_client_v1beta1.ModelServiceClient,
    prediction_service_client_v1beta1.PredictionServiceClient,
    pipeline_service_client_v1beta1.PipelineServiceClient,
    job_service_client_v1beta1.JobServiceClient,
    metadata_service_client_v1beta1.MetadataServiceClient,
    tensorboard_service_client_v1beta1.TensorboardServiceClient,
    vizier_service_client_v1beta1.VizierServiceClient,
    # v1
    dataset_service_client_v1.DatasetServiceClient,
    endpoint_service_client_v1.EndpointServiceClient,
    featurestore_online_serving_service_client_v1.FeaturestoreOnlineServingServiceClient,
    featurestore_service_client_v1.FeaturestoreServiceClient,
    metadata_service_client_v1.MetadataServiceClient,
    model_service_client_v1.ModelServiceClient,
    prediction_service_client_v1.PredictionServiceClient,
    pipeline_service_client_v1.PipelineServiceClient,
    job_service_client_v1.JobServiceClient,
    tensorboard_service_client_v1.TensorboardServiceClient,
    vizier_service_client_v1.VizierServiceClient,
)


RESOURCE_ID_PATTERN = re.compile(r"^[\w-]+$")


def validate_id(resource_id: str):
    """Validate resource ID.

    Args:
        resource_id (str): Resource id.
    Raises:
        ValueError: If resource id is not a valid format.

    """
    if not RESOURCE_ID_PATTERN.match(resource_id):
        raise ValueError(f"Resource {resource_id} is not a valid resource id.")


def full_resource_name(
    resource_name: str,
    resource_noun: str,
    parse_resource_name_method: Callable[[str], Dict[str, str]],
    format_resource_name_method: Callable[..., str],
    parent_resource_name_fields: Optional[Dict[str, str]] = None,
    project: Optional[str] = None,
    location: Optional[str] = None,
    resource_id_validator: Optional[Callable[[str], None]] = None,
) -> str:
    """Returns fully qualified resource name.

    Args:
        resource_name (str):
            Required. A fully-qualified Vertex AI resource name or
            resource ID.
        resource_noun (str):
            Required. A resource noun to validate the resource name against.
            For example, you would pass "datasets" to validate
            "projects/123/locations/us-central1/datasets/456".
        parse_resource_name_method (Callable[[str], Dict[str,str]]):
            Required. Method that parses a resource name into its segment parts.
            These are generally included with GAPIC clients.
        format_resource_name_method (Callable[..., str]):
            Required. Method that takes segment parts of resource names and returns
            the formated resource name. These are generally included with GAPIC clients.
        parent_resource_name_fields (Dict[str, str]):
            Optional. Dictionary of segment parts where key is the resource noun and
            values are the resource ids.
            For example:
                {
                    "metadataStores": "123"
                }
        project (str):
            Optional. project to retrieve resource_noun from. If not set, project
            set in aiplatform.init will be used.
        location (str):
            Optional. location to retrieve resource_noun from. If not set, location
            set in aiplatform.init will be used.
        resource_id_validator (Callable[str, None]):
            Optional. Function that validates the resource ID. Overrides the default validator, validate_id.
            Should take a resource ID as string and raise ValueError if invalid.

    Returns:
        resource_name (str):
            A fully-qualified Vertex AI resource name.
    """
    # Fully qualified resource name, e.g., "projects/.../locations/.../datasets/12345" or
    # "projects/.../locations/.../metadataStores/.../contexts/12345"
    fields = parse_resource_name_method(resource_name)
    if fields:
        return resource_name

    resource_id_validator = resource_id_validator or validate_id

    user_project = project or initializer.global_config.project
    user_location = location or initializer.global_config.location

    validate_region(user_location)
    resource_id_validator(resource_name)

    format_args = {
        "location": user_location,
        "project": user_project,
        convert_camel_case_resource_noun_to_snake_case(resource_noun): resource_name,
    }

    if parent_resource_name_fields:
        format_args.update(
            {
                convert_camel_case_resource_noun_to_snake_case(key): value
                for key, value in parent_resource_name_fields.items()
            }
        )

    return format_resource_name_method(**format_args)


# Resource nouns that are not plural in their resource names.
# Userd below to avoid conversion from plural to singular.
_SINGULAR_RESOURCE_NOUNS = {"time_series"}
_SINGULAR_RESOURCE_NOUNS_MAP = {"indexes": "index"}


def convert_camel_case_resource_noun_to_snake_case(resource_noun: str) -> str:
    """Converts camel case to snake case to map resource name parts to GAPIC parameter names.

    Args:
        resource_noun (str): The resource noun in camel case to covert.
    Returns:
        Singular snake case resource noun.
    """
    snake_case = re.sub("([A-Z]+)", r"_\1", resource_noun).lower()

    # plural to singular
    if snake_case in _SINGULAR_RESOURCE_NOUNS or not snake_case.endswith("s"):
        return snake_case
    elif snake_case in _SINGULAR_RESOURCE_NOUNS_MAP:
        return _SINGULAR_RESOURCE_NOUNS_MAP[snake_case]
    else:
        return snake_case[:-1]


def validate_display_name(display_name: str):
    """Verify display name is at most 128 chars.

    Args:
        display_name: display name to verify
    Raises:
        ValueError: display name is longer than 128 characters
    """
    if len(display_name) > 128:
        raise ValueError("Display name needs to be less than 128 characters.")


def validate_labels(labels: Dict[str, str]):
    """Validate labels.

    Args:
        labels: labels to verify
    Raises:
        ValueError: if labels is not a mapping of string key value pairs.
    """
    for k, v in labels.items():
        if not isinstance(k, str) or not isinstance(v, str):
            raise ValueError(
                "Expect labels to be a mapping of string key value pairs. "
                'Got "{}".'.format(labels)
            )


def validate_region(region: str) -> bool:
    """Validates region against supported regions.

    Args:
        region: region to validate
    Returns:
        bool: True if no errors raised
    Raises:
        ValueError: If region is not in supported regions.
    """
    if not region:
        raise ValueError(
            f"Please provide a region, select from {constants.SUPPORTED_REGIONS}"
        )

    region = region.lower()
    if region not in constants.SUPPORTED_REGIONS:
        raise ValueError(
            f"Unsupported region for Vertex AI, select from {constants.SUPPORTED_REGIONS}"
        )

    return True


def validate_accelerator_type(accelerator_type: str) -> bool:
    """Validates user provided accelerator_type string for training and
    prediction.

    Args:
        accelerator_type (str):
            Represents a hardware accelerator type.
    Returns:
        bool: True if valid accelerator_type
    Raises:
        ValueError if accelerator type is invalid.
    """
    if accelerator_type not in gca_accelerator_type.AcceleratorType._member_names_:
        raise ValueError(
            f"Given accelerator_type `{accelerator_type}` invalid. "
            f"Choose one of {gca_accelerator_type.AcceleratorType._member_names_}"
        )
    return True


def extract_bucket_and_prefix_from_gcs_path(gcs_path: str) -> Tuple[str, Optional[str]]:
    """Given a complete GCS path, return the bucket name and prefix as a tuple.

    Example Usage:

        bucket, prefix = extract_bucket_and_prefix_from_gcs_path(
            "gs://example-bucket/path/to/folder"
        )

        # bucket = "example-bucket"
        # prefix = "path/to/folder"

    Args:
        gcs_path (str):
            Required. A full path to a Google Cloud Storage folder or resource.
            Can optionally include "gs://" prefix or end in a trailing slash "/".

    Returns:
        Tuple[str, Optional[str]]
            A (bucket, prefix) pair from provided GCS path. If a prefix is not
            present, a None will be returned in its place.
    """
    if gcs_path.startswith("gs://"):
        gcs_path = gcs_path[5:]
    if gcs_path.endswith("/"):
        gcs_path = gcs_path[:-1]

    gcs_parts = gcs_path.split("/", 1)
    gcs_bucket = gcs_parts[0]
    gcs_blob_prefix = None if len(gcs_parts) == 1 else gcs_parts[1]

    return (gcs_bucket, gcs_blob_prefix)


def extract_project_and_location_from_parent(
    parent: str,
) -> Dict[str, str]:
    """Given a complete parent resource name, return the project and location as a dict.

    Example Usage:

        parent_resources = extract_project_and_location_from_parent(
            "projects/123/locations/us-central1/datasets/456"
        )

        parent_resources["project"] = "123"
        parent_resources["location"] = "us-central1"

    Args:
        parent (str):
            Required. A complete parent resource name.

    Returns:
        Dict[str, str]
            A project, location dict from provided parent resource name.
    """
    parent_resources = re.match(
        r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)(/|$)", parent
    )
    return parent_resources.groupdict() if parent_resources else {}


class ClientWithOverride:
    class WrappedClient:
        """Wrapper class for client that creates client at API invocation
        time."""

        def __init__(
            self,
            client_class: Type[VertexAiServiceClient],
            client_options: client_options.ClientOptions,
            client_info: gapic_v1.client_info.ClientInfo,
            credentials: Optional[auth_credentials.Credentials] = None,
        ):
            """Stores parameters needed to instantiate client.

            Args:
                client_class (VertexAiServiceClient):
                    Required. Class of the client to use.
                client_options (client_options.ClientOptions):
                    Required. Client options to pass to client.
                client_info (gapic_v1.client_info.ClientInfo):
                    Required. Client info to pass to client.
                credentials (auth_credentials.credentials):
                    Optional. Client credentials to pass to client.
            """

            self._client_class = client_class
            self._credentials = credentials
            self._client_options = client_options
            self._client_info = client_info

        def __getattr__(self, name: str) -> Any:
            """Instantiates client and returns attribute of the client."""
            temporary_client = self._client_class(
                credentials=self._credentials,
                client_options=self._client_options,
                client_info=self._client_info,
            )
            return getattr(temporary_client, name)

    @property
    @abc.abstractmethod
    def _is_temporary(self) -> bool:
        pass

    @property
    @classmethod
    @abc.abstractmethod
    def _default_version(self) -> str:
        pass

    @property
    @classmethod
    @abc.abstractmethod
    def _version_map(self) -> Tuple:
        pass

    @property
    def api_endpoint(self) -> str:
        """Default API endpoint used by this client."""
        client = self._clients[self._default_version]

        if self._is_temporary:
            return client._client_options.api_endpoint
        else:
            return client._transport._host.split(":")[0]

    def __init__(
        self,
        client_options: client_options.ClientOptions,
        client_info: gapic_v1.client_info.ClientInfo,
        credentials: Optional[auth_credentials.Credentials] = None,
    ):
        """Stores parameters needed to instantiate client.

        Args:
            client_options (client_options.ClientOptions):
                Required. Client options to pass to client.
            client_info (gapic_v1.client_info.ClientInfo):
                Required. Client info to pass to client.
            credentials (auth_credentials.credentials):
                Optional. Client credentials to pass to client.
        """

        self._clients = {
            version: self.WrappedClient(
                client_class=client_class,
                client_options=client_options,
                client_info=client_info,
                credentials=credentials,
            )
            if self._is_temporary
            else client_class(
                client_options=client_options,
                client_info=client_info,
                credentials=credentials,
            )
            for version, client_class in self._version_map
        }

    def __getattr__(self, name: str) -> Any:
        """Instantiates client and returns attribute of the client."""
        return getattr(self._clients[self._default_version], name)

    def select_version(self, version: str) -> VertexAiServiceClient:
        return self._clients[version]

    @classmethod
    def get_gapic_client_class(
        cls, version: Optional[str] = None
    ) -> Type[VertexAiServiceClient]:
        """Gets the underyilng GAPIC client.

        Used to access class and static methods without instantiating.

        Args:
            version (str):
                Optional. Version of client to retreive otherwise the default version is returned.
        Retuns:
            Underlying GAPIC client for this wrapper and version.
        """
        return dict(cls._version_map)[version or cls._default_version]


class DatasetClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, dataset_service_client_v1.DatasetServiceClient),
        (compat.V1BETA1, dataset_service_client_v1beta1.DatasetServiceClient),
    )


class EndpointClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, endpoint_service_client_v1.EndpointServiceClient),
        (compat.V1BETA1, endpoint_service_client_v1beta1.EndpointServiceClient),
    )


class IndexClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, index_service_client_v1.IndexServiceClient),
        (compat.V1BETA1, index_service_client_v1beta1.IndexServiceClient),
    )


class IndexEndpointClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, index_endpoint_service_client_v1.IndexEndpointServiceClient),
        (
            compat.V1BETA1,
            index_endpoint_service_client_v1beta1.IndexEndpointServiceClient,
        ),
    )


class FeaturestoreClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, featurestore_service_client_v1.FeaturestoreServiceClient),
        (compat.V1BETA1, featurestore_service_client_v1beta1.FeaturestoreServiceClient),
    )


class FeaturestoreOnlineServingClientWithOverride(ClientWithOverride):
    _is_temporary = False
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (
            compat.V1,
            featurestore_online_serving_service_client_v1.FeaturestoreOnlineServingServiceClient,
        ),
        (
            compat.V1BETA1,
            featurestore_online_serving_service_client_v1beta1.FeaturestoreOnlineServingServiceClient,
        ),
    )


class JobClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, job_service_client_v1.JobServiceClient),
        (compat.V1BETA1, job_service_client_v1beta1.JobServiceClient),
    )


class ModelClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, model_service_client_v1.ModelServiceClient),
        (compat.V1BETA1, model_service_client_v1beta1.ModelServiceClient),
    )


class PipelineClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, pipeline_service_client_v1.PipelineServiceClient),
        (compat.V1BETA1, pipeline_service_client_v1beta1.PipelineServiceClient),
    )


class PipelineJobClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, pipeline_service_client_v1.PipelineServiceClient),
        (compat.V1BETA1, pipeline_service_client_v1beta1.PipelineServiceClient),
    )


class PredictionClientWithOverride(ClientWithOverride):
    _is_temporary = False
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, prediction_service_client_v1.PredictionServiceClient),
        (compat.V1BETA1, prediction_service_client_v1beta1.PredictionServiceClient),
    )


class MetadataClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, metadata_service_client_v1.MetadataServiceClient),
        (compat.V1BETA1, metadata_service_client_v1beta1.MetadataServiceClient),
    )


class TensorboardClientWithOverride(ClientWithOverride):
    _is_temporary = False
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, tensorboard_service_client_v1.TensorboardServiceClient),
        (compat.V1BETA1, tensorboard_service_client_v1beta1.TensorboardServiceClient),
    )


class VizierClientWithOverride(ClientWithOverride):
    _is_temporary = True
    _default_version = compat.DEFAULT_VERSION
    _version_map = (
        (compat.V1, vizier_service_client_v1.VizierServiceClient),
        (compat.V1BETA1, vizier_service_client_v1beta1.VizierServiceClient),
    )


VertexAiServiceClientWithOverride = TypeVar(
    "VertexAiServiceClientWithOverride",
    DatasetClientWithOverride,
    EndpointClientWithOverride,
    FeaturestoreClientWithOverride,
    JobClientWithOverride,
    ModelClientWithOverride,
    PipelineClientWithOverride,
    PipelineJobClientWithOverride,
    PredictionClientWithOverride,
    MetadataClientWithOverride,
    TensorboardClientWithOverride,
    VizierClientWithOverride,
)


class LoggingFilter(logging.Filter):
    def __init__(self, warning_level: int):
        self._warning_level = warning_level

    def filter(self, record):
        return record.levelname == self._warning_level


def _timestamped_gcs_dir(root_gcs_path: str, dir_name_prefix: str) -> str:
    """Composes a timestamped GCS directory.

    Args:
        root_gcs_path: GCS path to put the timestamped directory.
        dir_name_prefix: Prefix to add the timestamped directory.
    Returns:
        Timestamped gcs directory path in root_gcs_path.
    """
    timestamp = datetime.datetime.now().isoformat(sep="-", timespec="milliseconds")
    dir_name = "-".join([dir_name_prefix, timestamp])
    if root_gcs_path.endswith("/"):
        root_gcs_path = root_gcs_path[:-1]
    gcs_path = "/".join([root_gcs_path, dir_name])
    if not gcs_path.startswith("gs://"):
        return "gs://" + gcs_path
    return gcs_path


def _timestamped_copy_to_gcs(
    local_file_path: str,
    gcs_dir: str,
    project: Optional[str] = None,
    credentials: Optional[auth_credentials.Credentials] = None,
) -> str:
    """Copies a local file to a GCS path.

    The file copied to GCS is the name of the local file prepended with an
    "aiplatform-{timestamp}-" string.

    Args:
        local_file_path (str): Required. Local file to copy to GCS.
        gcs_dir (str):
            Required. The GCS directory to copy to.
        project (str):
            Project that contains the staging bucket. Default will be used if not
            provided. Model Builder callers should pass this in.
        credentials (auth_credentials.Credentials):
            Custom credentials to use with bucket. Model Builder callers should pass
            this in.
    Returns:
        gcs_path (str): The path of the copied file in gcs.
    """

    gcs_bucket, gcs_blob_prefix = extract_bucket_and_prefix_from_gcs_path(gcs_dir)

    local_file_name = pathlib.Path(local_file_path).name
    timestamp = datetime.datetime.now().isoformat(sep="-", timespec="milliseconds")
    blob_path = "-".join(["aiplatform", timestamp, local_file_name])

    if gcs_blob_prefix:
        blob_path = "/".join([gcs_blob_prefix, blob_path])

    # TODO(b/171202993) add user agent
    client = storage.Client(project=project, credentials=credentials)
    bucket = client.bucket(gcs_bucket)
    blob = bucket.blob(blob_path)
    blob.upload_from_filename(local_file_path)

    gcs_path = "".join(["gs://", "/".join([blob.bucket.name, blob.name])])
    return gcs_path


def get_timestamp_proto(
    time: Optional[datetime.datetime] = None,
) -> timestamp_pb2.Timestamp:
    """Gets timestamp proto of a given time.
    Args:
        time (datetime.datetime):
            Optional. A user provided time. Default to datetime.datetime.now() if not given.
    Returns:
        timestamp_pb2.Timestamp: timestamp proto of the given time, not have higher than millisecond precision.
    """
    if not time:
        time = datetime.datetime.now()

    time_str = time.isoformat(sep=" ", timespec="milliseconds")
    time = datetime.datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S.%f")

    timestamp_proto = timestamp_pb2.Timestamp()
    timestamp_proto.FromDatetime(time)

    return timestamp_proto


def timestamped_unique_name() -> str:
    """Composes a timestamped unique name.

    Returns:
        A string representing a unique name.
    """
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    unique_id = uuid.uuid4().hex[0:5]
    return f"{timestamp}-{unique_id}"



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1.schema.trainingjob.definition",
    manifest={
        "ExportEvaluatedDataItemsConfig",
    },
)


class ExportEvaluatedDataItemsConfig(proto.Message):
    r"""Configuration for exporting test set predictions to a
    BigQuery table.

    Attributes:
        destination_bigquery_uri (str):
            URI of desired destination BigQuery table. Expected format:
            bq://<project_id>:<dataset_id>:

            If not specified, then results are exported to the following
            auto-created BigQuery table:
            <project_id>:export_evaluated_examples_<model_name>_<yyyy_MM_dd'T'HH_mm_ss_SSS'Z'>.evaluated_examples
        override_existing_table (bool):
            If true and an export destination is
            specified, then the contents of the destination
            are overwritten. Otherwise, if the export
            destination already exists, then the export
            operation fails.
    """

    destination_bigquery_uri: str = proto.Field(
        proto.STRING,
        number=1,
    )
    override_existing_table: bool = proto.Field(
        proto.BOOL,
        number=2,
    )


__all__ = tuple(sorted(__protobuf__.manifest))
